/****************************************************************************
** Meta object code from reading C++ file 'dfilefuture.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../../util-dfm/include/dfm-io/dfm-io/dfilefuture.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'dfilefuture.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN5dfmio11DFileFutureE_t {};
} // unnamed namespace

template <> constexpr inline auto dfmio::DFileFuture::qt_create_metaobjectdata<qt_meta_tag_ZN5dfmio11DFileFutureE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "dfmio::DFileFuture",
        "finished",
        "",
        "infoAttribute",
        "DFileInfo::AttributeID",
        "id",
        "QVariant",
        "value",
        "key",
        "progressNotify",
        "current",
        "total",
        "infoExists",
        "exists",
        "infoPermissions",
        "DFile::Permissions",
        "permissions",
        "infoSize",
        "size",
        "writeAsyncSize",
        "readData",
        "data",
        "infoMedia",
        "QUrl",
        "url",
        "QMap<DFileInfo::AttributeExtendID,QVariant>",
        "map"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'finished'
        QtMocHelpers::SignalData<void()>(1, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'infoAttribute'
        QtMocHelpers::SignalData<void(DFileInfo::AttributeID, const QVariant &)>(3, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 4, 5 }, { 0x80000000 | 6, 7 },
        }}),
        // Signal 'infoAttribute'
        QtMocHelpers::SignalData<void(const QByteArray &, const QVariant &)>(3, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QByteArray, 8 }, { 0x80000000 | 6, 7 },
        }}),
        // Signal 'progressNotify'
        QtMocHelpers::SignalData<void(qint64, qint64)>(9, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::LongLong, 10 }, { QMetaType::LongLong, 11 },
        }}),
        // Signal 'infoExists'
        QtMocHelpers::SignalData<void(const bool)>(12, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 13 },
        }}),
        // Signal 'infoPermissions'
        QtMocHelpers::SignalData<void(const DFile::Permissions)>(14, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 15, 16 },
        }}),
        // Signal 'infoSize'
        QtMocHelpers::SignalData<void(const quint64 &)>(17, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::ULongLong, 18 },
        }}),
        // Signal 'writeAsyncSize'
        QtMocHelpers::SignalData<void(const qint64 &)>(19, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::LongLong, 18 },
        }}),
        // Signal 'readData'
        QtMocHelpers::SignalData<void(const QByteArray &)>(20, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QByteArray, 21 },
        }}),
        // Signal 'infoMedia'
        QtMocHelpers::SignalData<void(const QUrl &, const QMap<DFileInfo::AttributeExtendID,QVariant> &)>(22, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 23, 24 }, { 0x80000000 | 25, 26 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<DFileFuture, qt_meta_tag_ZN5dfmio11DFileFutureE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject dfmio::DFileFuture::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN5dfmio11DFileFutureE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN5dfmio11DFileFutureE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN5dfmio11DFileFutureE_t>.metaTypes,
    nullptr
} };

void dfmio::DFileFuture::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<DFileFuture *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->finished(); break;
        case 1: _t->infoAttribute((*reinterpret_cast<std::add_pointer_t<DFileInfo::AttributeID>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QVariant>>(_a[2]))); break;
        case 2: _t->infoAttribute((*reinterpret_cast<std::add_pointer_t<QByteArray>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QVariant>>(_a[2]))); break;
        case 3: _t->progressNotify((*reinterpret_cast<std::add_pointer_t<qint64>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<qint64>>(_a[2]))); break;
        case 4: _t->infoExists((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 5: _t->infoPermissions((*reinterpret_cast<std::add_pointer_t<DFile::Permissions>>(_a[1]))); break;
        case 6: _t->infoSize((*reinterpret_cast<std::add_pointer_t<quint64>>(_a[1]))); break;
        case 7: _t->writeAsyncSize((*reinterpret_cast<std::add_pointer_t<qint64>>(_a[1]))); break;
        case 8: _t->readData((*reinterpret_cast<std::add_pointer_t<QByteArray>>(_a[1]))); break;
        case 9: _t->infoMedia((*reinterpret_cast<std::add_pointer_t<QUrl>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QMap<DFileInfo::AttributeExtendID,QVariant>>>(_a[2]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (DFileFuture::*)()>(_a, &DFileFuture::finished, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (DFileFuture::*)(DFileInfo::AttributeID , const QVariant & )>(_a, &DFileFuture::infoAttribute, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (DFileFuture::*)(const QByteArray & , const QVariant & )>(_a, &DFileFuture::infoAttribute, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (DFileFuture::*)(qint64 , qint64 )>(_a, &DFileFuture::progressNotify, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (DFileFuture::*)(const bool )>(_a, &DFileFuture::infoExists, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (DFileFuture::*)(const DFile::Permissions )>(_a, &DFileFuture::infoPermissions, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (DFileFuture::*)(const quint64 & )>(_a, &DFileFuture::infoSize, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (DFileFuture::*)(const qint64 & )>(_a, &DFileFuture::writeAsyncSize, 7))
            return;
        if (QtMocHelpers::indexOfMethod<void (DFileFuture::*)(const QByteArray & )>(_a, &DFileFuture::readData, 8))
            return;
        if (QtMocHelpers::indexOfMethod<void (DFileFuture::*)(const QUrl & , const QMap<DFileInfo::AttributeExtendID,QVariant> & )>(_a, &DFileFuture::infoMedia, 9))
            return;
    }
}

const QMetaObject *dfmio::DFileFuture::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *dfmio::DFileFuture::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN5dfmio11DFileFutureE_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int dfmio::DFileFuture::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 10)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 10;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 10)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 10;
    }
    return _id;
}

// SIGNAL 0
void dfmio::DFileFuture::finished()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void dfmio::DFileFuture::infoAttribute(DFileInfo::AttributeID _t1, const QVariant & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1, _t2);
}

// SIGNAL 2
void dfmio::DFileFuture::infoAttribute(const QByteArray & _t1, const QVariant & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1, _t2);
}

// SIGNAL 3
void dfmio::DFileFuture::progressNotify(qint64 _t1, qint64 _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1, _t2);
}

// SIGNAL 4
void dfmio::DFileFuture::infoExists(const bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 4, nullptr, _t1);
}

// SIGNAL 5
void dfmio::DFileFuture::infoPermissions(const DFile::Permissions _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 5, nullptr, _t1);
}

// SIGNAL 6
void dfmio::DFileFuture::infoSize(const quint64 & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 6, nullptr, _t1);
}

// SIGNAL 7
void dfmio::DFileFuture::writeAsyncSize(const qint64 & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 7, nullptr, _t1);
}

// SIGNAL 8
void dfmio::DFileFuture::readData(const QByteArray & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 8, nullptr, _t1);
}

// SIGNAL 9
void dfmio::DFileFuture::infoMedia(const QUrl & _t1, const QMap<DFileInfo::AttributeExtendID,QVariant> & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 9, nullptr, _t1, _t2);
}
QT_WARNING_POP
